"""
Package: src.tests
Filename: test_helpers.py
Author(s): Grant W

Description: Test our helpers
"""
# Python Imports
from unittest.mock import patch
import random
import time

# Third Party Imports
import requests
import pytest

# Dynata Imports
import dynata_rex

# Local Imports


def test_make_session():
    """Test the make_session helper"""

    session = dynata_rex.helpers.make_session()

    assert isinstance(session, requests.Session)

    assert session.headers["User-Agent"] == 'rex-sdk-python/0.0.1'

    assert isinstance(session.adapters['https://'],
                      dynata_rex.helpers.TimeoutHTTPAdapter)
    assert isinstance(session.adapters['http://'],
                      dynata_rex.helpers.TimeoutHTTPAdapter)


@patch.object(requests.adapters.HTTPAdapter, 'send')
def test_timeout_adapter_raises_timeout_exception(adapter_send):
    session = dynata_rex.helpers.make_session()
    with pytest.raises(dynata_rex.exceptions.HttpTimeoutException):
        adapter_send.side_effect = requests.exceptions.ReadTimeout()
        session.get('http://example.com')
        session.get('https://example.com')


@patch.object(dynata_rex.helpers.TimeoutHTTPAdapter, '_wait')
@patch.object(requests.adapters.HTTPAdapter, 'send')
def test_timeout_adapter_raises_rex_exception(adapter_send, wait_method):
    session = dynata_rex.helpers.make_session()
    adapter_send.side_effect = requests.exceptions.ConnectionError()
    with pytest.raises(dynata_rex.exceptions.RexServiceException):
        session.get('https://example.com')
        assert wait_method.call_count == dynata_rex.helpers.DEFAULT_RETRIES


@patch.object(time, 'sleep')
@patch.object(random, 'randint')
def test_wait_method(randint_func, sleep_func):
    randint_func.return_value = 1000
    sleep_func.return_value = None

    adapter = dynata_rex.helpers.TimeoutHTTPAdapter(request_timeout=1)

    adapter._wait(1)
    assert sleep_func.call_count == 1

    # should sleep 3 seconds for 1st try, where randint returns 1000
    assert sleep_func.call_args[0][0] == 3


@patch.object(requests.adapters.HTTPAdapter, 'send')
def test_request_timeout_in_kwargs(parent_send_adapter):
    """Test that the request_timeout is passed to the adapter"""
    default_timeout = 10
    desired_timeout_for_request = 20

    adapter = dynata_rex.helpers.TimeoutHTTPAdapter(
        request_timeout=default_timeout)

    adapter.send(requests.Request('GET',
                 'http://example.com'),
                 request_timeout=desired_timeout_for_request)

    assert parent_send_adapter.call_args[1]['timeout'] == \
        desired_timeout_for_request



"""
Package: dynata-rex.models
Filename: respondent_gateway.py
Author(s): Grant W

Description: Implementation of dataclasses for Respondent Gateway
"""
# Python Imports
from enum import Enum
from typing import Optional


class GatewayGenderEnum(Enum):
    MALE = 1
    FEMALE = 2


class GatewayDispositionsEnum(Enum):
    UNKNOWN = 0
    COMPLETE = 1
    TERMINATION = 2
    OVERQUOTA = 3
    DUPLICATE = 4
    QUALITY = 5


class GatewayStatusEnum(Enum):
    """
    Composite status code from Disposition + non-unique status code
    Must call as tuple of Gateway Disposition + Status code
    ie
    >>> disp_code, status_code = 1, 0
    >>> disp = GatewayDispositionsEnum(disp_code)
    >>> status = RespondentGatewayStatusEnum((disp, status_code))
    >>> status.name
    'COMPLETE_DEFAULT`
    >>> status.value
    (<GatewayDispositionsEnum.COMPLETE: 1>, 0)
    """
    UNKNOWN_DEFAULT = GatewayDispositionsEnum.UNKNOWN, 0
    COMPLETE_DEFAULT = GatewayDispositionsEnum.COMPLETE, 0
    COMPLETE_PARTIAL = GatewayDispositionsEnum.COMPLETE, 1
    TERMINATION_DYNATA = GatewayDispositionsEnum.TERMINATION, 1
    TERMINATION_CLIENT = GatewayDispositionsEnum.TERMINATION, 2
    OVERQUOTA_DYNATA = GatewayDispositionsEnum.OVERQUOTA, 1
    OVERQUOTA_CLIENT = GatewayDispositionsEnum.OVERQUOTA, 2
    DUPLICATE_DEFAULT = GatewayDispositionsEnum.DUPLICATE, 0
    QUALITY_ANSWER = GatewayDispositionsEnum.QUALITY, 1
    QUALITY_SPEEDING = GatewayDispositionsEnum.QUALITY, 2
    QUALITY_SUSPENDED = GatewayDispositionsEnum.QUALITY, 3


class Attribute:

    def __init__(self, attribute_id: int, answers: list[int]):
        self.id = attribute_id
        self.answers = answers

    def __iter__(self):
        yield from {
            "id": self.id,
            "answers": self.answers
        }.items()

    def __str__(self):
        return str(dict(self))

    def to_json(self):
        return dict(self)


class PutRespondentRequest:
    def __init__(self, respondent_id: str,
                 language: str,
                 country: str,
                 gender: Optional[str],
                 birth_date: Optional[str],
                 postal_code: Optional[str],
                 attributes: Optional[list[Attribute]]):
        self.respondent_id = respondent_id
        self.language = language
        self.country = country
        self.gender = gender
        self.birth_date = birth_date
        self.postal_code = postal_code
        self.attributes = attributes

    def __iter__(self):
        yield from {
            "respondent_id": self.respondent_id,
            "language": self.language,
            "country": self.country,
            "gender": self.gender,
            "birth_date": self.birth_date,
            "postal_code": self.postal_code,
            "attributes": [
                attribute.to_json() for attribute in self.attributes
            ],
        }.items()

    def __str__(self):
        return str(dict(self))

    def to_json(self):
        return dict(self)


class PutRespondentAnswersRequest:
    def __init__(self,
                 respondent_id: str,
                 attributes: list[Attribute]):
        self.respondent_id = respondent_id
        self.attributes = attributes

    def __iter__(self):
        yield from {
            "respondent_id": self.respondent_id,
            "attributes": [
                attribute.to_json() for attribute in self.attributes
            ]
        }.items()

    def __str__(self):
        return str(dict(self))

    def to_json(self):
        return dict(self)



"""
Package: src.dynata_rex
Filename: respondent_gateway.py
Author(s): Grant W

Description: Respondent Gateway interactions
"""
# Python Import
from urllib.parse import urlparse, parse_qsl, unquote, urlencode
from typing import Union
from copy import copy

# Third Party Imports

# Local Imports
from dynata_rex.models import GatewayDispositionsEnum, \
    GatewayStatusEnum, \
    PutRespondentRequest, \
    PutRespondentAnswersRequest
from .signer import Signer, RexRequest
from .exceptions import SignatureExpiredException, SignatureInvalidException


class RespondentGateway:
    """
    Respondent Gateway interactions
    """

    _BASE_URL = 'https://respondent.rex.dynata.com'

    def __init__(self,
                 access_key: str,
                 secret_key: str,
                 base_url: str = _BASE_URL,
                 default_ttl: int = 10):
        """
        @access_key: liam access key for REX
        @secret_key: liam secret key for REX

        # Optional
        @base_url: url of Gateway
        @ttl: time to live for signature in seconds
        """
        self.access_key = access_key
        self.secret_key = secret_key
        self.default_ttl = default_ttl
        self.base_url = base_url

        # MR for API requests
        self.make_request = RexRequest(access_key,
                                       secret_key,
                                       default_ttl=default_ttl)
        # Signer for signing/verifying URLs
        self.signer = Signer(access_key, secret_key, default_ttl=default_ttl)

    def create_respondent_url(self,
                              url: str,
                              birth_date: str,
                              gender: str,
                              postal_code: str,
                              respondent_id: str,
                              additional_params: dict = None,
                              ttl: Union[int, None] = None,
                              url_quoting: bool = False) -> str:
        """
        Creates a url from the url of an opportunity + the required
        parameters for entry into the respondent gateway.

        @url: live/test url of an opportunity
        @birth_date: birth date in ISO8601 (YYYY-MM-DD) format
        @gender: 'male' / 'female'
        @postal: postal code as string
        @respondent_id : unique identifier for the respondent

        @additional_params: additional parameters that will be included
            on the return back from the survey
        @ttl: time to live for signature in seconds
        @url_quoting: whether to URL quote the returned URL
        """
        parsed = urlparse(url)
        base_params = dict(parse_qsl(parsed.query))
        if additional_params:
            base_params = dict(base_params, **additional_params)
        required_params = {
            'birth_date': birth_date,
            'gender': gender,
            'postal_code': postal_code,
            'respondent_id': respondent_id
        }
        final_params = dict(base_params, **required_params)
        updated_query = urlencode(final_params, doseq=True)
        parsed = parsed._replace(query=updated_query)
        return self.sign_url(parsed.geturl(), ttl=ttl, url_quoting=url_quoting)

    def sign_url(self,
                 url,
                 ttl: Union[int, None] = None,
                 url_quoting: bool = False) -> str:
        """
        Sign a URL with the given access and secret keys

        @url: URL to sign
        @ttl: time to live for signature in seconds
        @url_quoting: whether to URL quote the returned URL
        """
        if not ttl:
            ttl = self.default_ttl
        parsed = urlparse(url)
        query_parameters = dict(parse_qsl(parsed.query))
        signed_params = self.signer.sign_query_parameters_from_ttl(
            query_parameters, ttl=ttl)
        updated = parsed._replace(query=signed_params).geturl()
        if url_quoting:
            return updated
        return unquote(updated)

    def verify_query_parameters(self,
                                query_parameters: dict,
                                access_key: str = None,
                                secret_key: str = None) -> bool:
        """
        Verify a signature on query parameters match the expected output
        from a pair of access and secret keys. Returns True if the signature
        matches, False otherwise.

        @query_parameters: dictionary of query parameters

        Optional
        @access_key: liam access key for signing
        @secret_key: liam secret key for signing
        """
        parameters = copy(query_parameters)
        if access_key is None:
            access_key = self.access_key
        if secret_key is None:
            secret_key = self.secret_key
        expiration_date_str = parameters.pop('expiration')
        if self.signer.is_expired(expiration_date_str):
            raise SignatureExpiredException
        original_signature = parameters.pop('signature')
        signed = self.signer.sign_query_params_from_expiration_date(
            parameters, expiration_date_str, access_key, secret_key,
            as_dict=True
        )
        if original_signature != signed['signature']:
            raise SignatureInvalidException
        return True

    def verify_url(self,
                   url,
                   access_key: str = None,
                   secret_key: str = None) -> bool:
        """
        Verify a URL's signature matches for the given access and secret keys
        @url: URL to verify

        Optional
        @access_key: liam access key for signing
        @secret_key: liam secret key for signing
        """
        if access_key is None:
            access_key = self.access_key
        if secret_key is None:
            secret_key = self.secret_key
        parsed = urlparse(url)
        query_parameters = dict(parse_qsl(parsed.query))
        try:
            return self.verify_query_parameters(query_parameters,
                                                access_key=access_key,
                                                secret_key=secret_key)
        except (KeyError,
                SignatureExpiredException,
                SignatureInvalidException):
            return False

    def get_respondent_disposition(
            self, url) -> Union[GatewayDispositionsEnum, None]:
        """
        Get the disposition of a respondent from a URL

        @url: URL to get disposition from
        """
        parsed = urlparse(url)
        query_parameters = dict(parse_qsl(parsed.query))
        try:
            disposition = int(query_parameters['disposition'])
            return GatewayDispositionsEnum(disposition)
        except KeyError:
            return None

    def get_respondent_status(
            self, url) -> Union[GatewayStatusEnum, None]:
        """
        Get the status of a respondent from a URL

        @url: URL to get status from
        """
        parsed = urlparse(url)
        query_parameters = dict(parse_qsl(parsed.query))
        disposition = self.get_respondent_disposition(url)
        if not disposition:
            return None
        try:
            status = int(query_parameters['status'])
            return GatewayStatusEnum((disposition, status))
        except KeyError:
            return None

    def create_context(self, context_id: str, context_data: dict) -> int:
        """
        Create a context with the given context_id and context_data

        @context_id: unique identifier for the context
        @context_data: dictionary of context data ie:
            {
                "ctx": "a987dsglh34t435jkhsdg98u",
                "gender": "male",
                "postal_code": "60081",
                "birth_date": "1959-10-05",
                "country": "US"
            }
        """
        endpoint = f"{self.base_url}/create-context"
        data = {
            "id": context_id,
            "items": context_data
        }
        response = self.make_request.post(endpoint, data)
        return response['id']

    def expire_context(self, context_id: str) -> None:
        """
        Expire a context with the given context_id and account_id

        @context_id: identifier for the context
        """
        endpoint = f"{self.base_url}/expire-context"
        data = {"id": context_id}
        res = self.make_request.post(endpoint, data)
        return res if res else None

    def get_context(self, context_id: int) -> dict:
        """
        Get specific opportunity from SMOR

        @context_id: identifier for the context
        """
        endpoint = f"{self.base_url}/get-context"
        data = {"id": context_id}
        return self.make_request.post(endpoint, data)

    def get_attribute_info(self, attribute_id: int) -> dict:
        """
        Get parameter info about a specific attribute id

        @attribute_id: specific singular id for a parameter item
        """
        endpoint = f"{self.base_url}/get-attribute-info"
        data = {"attribute_id": attribute_id}
        return self.make_request.post(endpoint, data)

    def get_attributes(self,
                       country: str,
                       page_number: int,
                       page_size: int) -> dict:
        """
        [Deprecated - please use list_attributes()]

        Get a list of attribute id's and their statuses

        @country: Country code for which you would like attributes

        @page_number: What page number you are requesting (for pagination)

        @page_size: How many id's you would like returned in each page
        """
        endpoint = f"{self.base_url}/get-attributes"
        data = {
            "country": country,
            "page_number": page_number,
            "page_size": page_size
        }
        return self.make_request.post(endpoint, data)

    def list_attributes(self,
                        country: str,
                        page_number: int,
                        page_size: int) -> dict:
        """
        Get a list of attribute id's and their statuses

        @country: Country code for which you would like attributes

        @page_number: What page number you are requesting (for pagination)

        @page_size: How many id's you would like returned in each page
        """
        endpoint = f"{self.base_url}/list-attributes"
        data = {
            "country": country,
            "page_number": page_number,
            "page_size": page_size
        }
        return self.make_request.post(endpoint, data)

    def put_respondent(self,
                       request: PutRespondentRequest
                       ) -> Union[dict, str]:
        """
        Create a respondent from its models.

        :param request: Respondent model
        :type request: PutRespondentRequest
        """
        endpoint = f"{self.base_url}/put-respondent"
        return self.make_request.post(endpoint, request.to_json())

    def put_respondent_answers(self,
                               request: PutRespondentAnswersRequest
                               ) -> Union[dict, str]:
        """
        Add answers to respondent.

        :param request: Respondent answers model
        :type request: PutRespondentAnswersRequest
        """
        endpoint = f"{self.base_url}/put-respondent-answers"
        return self.make_request.post(endpoint, request.to_json())



"""
Package: src.tests
Filename: test_gateway.py
Author(s): Gino T

Description: Test our requester

NOTE: Requester mainly uses a TTL (seconds from current time), so we have to
mock the expiration date generated quite a bit in here
"""
# Python Imports
from unittest.mock import patch
from urllib.parse import urlparse, parse_qsl
from pytest import raises
import json

# Third Party Imports
import requests

from dynata_rex.models.respondent_gateway import Attribute
# Dynata Imports
from dynata_rex.respondent_gateway import RespondentGateway
from dynata_rex.signer import Signer, RexRequest
from dynata_rex.models import (GatewayDispositionsEnum,
                               GatewayStatusEnum,
                               PutRespondentAnswersRequest)
from dynata_rex.exceptions import (SignatureExpiredException,
                                   SignatureInvalidException)
# Local Imports
from .shared import (ACCESS_KEY,
                     SECRET_KEY,
                     TEST_DATE_STR,
                     ResponseMock,
                     TEST_DATA)

GATEWAY = RespondentGateway(
    ACCESS_KEY,
    SECRET_KEY,
    "https://respondent.rex.dynata.com"
)
REQUESTER = RexRequest(ACCESS_KEY, SECRET_KEY)


@patch.object(Signer, "create_expiration_date")
def test_create_respondent_url(fun):
    expected = "https://respondent.qa-rex.dynata.com/start" \
               "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
               "&language=es" \
               "&birth_date=1989-09-16" \
               "&gender=male&postal_code=00000" \
               "&respondent_id=12345" \
               "&access_key=access_key" \
               "&expiration=1970-01-01T00:00:00.000Z" \
               "&signature=3d063d93efde81823ca0e" \
               "224c2458cc6053da0f0ce2ff5833644c8e5eddbc594"

    # Mock return from create_expiration_date()
    fun.return_value = TEST_DATE_STR

    signed_url = GATEWAY.create_respondent_url(
        'https://respondent.qa-rex.dynata.com/start'
        '?ctx=7c26bf58'
        '-43db-4370-977d-d14fa4356930&language=es',
        '1989-09-16',
        'male',
        '00000',
        '12345')
    assert expected == signed_url


@patch.object(Signer, "create_expiration_date")
def test_create_respondent_url_additional_params(fun):
    expected = "https://respondent.qa-rex.dynata.com/start" \
               "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
               "&language=es" \
               "&param1=popcorn" \
               "&birth_date=1989-09-16" \
               "&gender=male" \
               "&postal_code=00000" \
               "&respondent_id=12345" \
               "&access_key=access_key" \
               "&expiration=1970-01-01T00:00:00.000Z" \
               "&signature=5c4c8748a1d91a1407f" \
               "7da76f4a5a24a3116b0bd94d9326722a5913ac99cb2a3"

    # Mock return from create_expiration_date()
    fun.return_value = TEST_DATE_STR

    signed_url = GATEWAY.create_respondent_url(
        'https://respondent.qa-rex.dynata.com/start?ctx=7c26bf58'
        '-43db-4370-977d-d14fa4356930&language=es',
        '1989-09-16',
        'male',
        '00000',
        '12345',
        {'param1': 'popcorn'})

    assert expected == signed_url


@patch.object(Signer, "create_expiration_date")
def test_create_respondent_url_url_quoting(fun):
    expected = "https://respondent.qa-rex.dynata.com/start" \
               "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
               "&language=es" \
               "&birth_date=1989-09-16" \
               "&gender=male&postal_code=00000" \
               "&respondent_id=12345" \
               "&access_key=access_key" \
               "&expiration=1970-01-01T00%3A00%3A00.000Z" \
               "&signature=3d063d93efde81823ca0e" \
               "224c2458cc6053da0f0ce2ff5833644c8e5eddbc594"

    # Mock return from create_expiration_date()
    fun.return_value = TEST_DATE_STR

    signed_url = GATEWAY.create_respondent_url(
        'https://respondent.qa-rex.dynata.com/start?ctx=7c26bf58'
        '-43db-4370-977d-d14fa4356930&language=es',
        '1989-09-16',
        'male',
        '00000',
        '12345',
        {},
        None, True)

    assert expected == signed_url


@patch.object(Signer, "create_expiration_date")
def test_sign_url_url_quota(fun):
    expected = "https://respondent.qa-rex.dynata.com/start" \
               "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
               "&language=es" \
               "&access_key=access_key" \
               "&expiration=1970-01-01T00%3A00%3A00.000Z" \
               "&signature=69f06a6d0f4efba81a6" \
               "2c1eca08decfb64acc0461ea994ce7ae24fb99f7f9e11"

    # Mock return from create_expiration_date()
    fun.return_value = TEST_DATE_STR

    signed_url = GATEWAY.sign_url(
        'https://respondent.qa-rex.dynata.com/start?ctx=7c26bf58'
        '-43db-4370-977d-d14fa4356930&language=es',
        None,
        True
    )
    assert expected == signed_url


@patch.object(Signer, "create_expiration_date")
def test_sign_url(fun):
    expected = "https://respondent.qa-rex.dynata.com/start" \
               "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
               "&language=es" \
               "&access_key=access_key" \
               "&expiration=1970-01-01T00:00:00.000Z" \
               "&signature=69f06a6d0f4efba81a6" \
               "2c1eca08decfb64acc0461ea994ce7ae24fb99f7f9e11"

    # Mock return from create_expiration_date()
    fun.return_value = "1970-01-01T00:00:00.000Z"

    signed_url = GATEWAY.sign_url(
        'https://respondent.qa-rex.dynata.com/start?ctx=7c26bf58'
        '-43db-4370-977d-d14fa4356930&language=es',
        None,
        False
    )
    assert expected == signed_url


@patch.object(Signer, "create_expiration_date")
def test_verify_query_params(fun):
    url = "https://respondent.qa-rex.dynata.com/start" \
          "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
          "&language=es" \
          "&access_key=access_key" \
          "&expiration=2099-01-01T00:00:00.000Z" \
          "&signature=e3bc08b50ac9aa2612558" \
          "28f741467a956c40c4ce1f0e29bb1bcd23b81f31959"

    # Mock return from create_expiration_date()
    fun.return_value = "2099-01-01T00:00:00.000Z"
    parsed = urlparse(url)
    query_parameters = dict(parse_qsl(parsed.query))

    assert GATEWAY.verify_query_parameters(query_parameters, 'access_key')


@patch.object(Signer, "create_expiration_date")
def test_verify_query_params_signature_expired(fun):
    with raises(SignatureExpiredException) as excinfo:
        url = "https://respondent.qa-rex.dynata.com/start" \
              "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
              "&language=es" \
              "&access_key=access_key" \
              "&expiration=2002-01-01T00:00:00.000Z" \
              "&signature=e3bc08b50ac9aa261255" \
              "828f741467a956c40c4ce1f0e29bb1bcd23b81f31959"

        # Mock return from create_expiration_date()
        fun.return_value = "2002-01-01T00:00:00.000Z"
        parsed = urlparse(url)
        query_parameters = dict(parse_qsl(parsed.query))
        GATEWAY.verify_query_parameters(query_parameters, 'access_key')
    assert excinfo.type == SignatureExpiredException


@patch.object(Signer, "create_expiration_date")
def test_verify_query_params_signature_invalid(fun):
    with raises(SignatureInvalidException) as excinfo:
        url = "https://respondent.qa-rex.dynata.com/start" \
              "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
              "&language=es" \
              "&access_key=access_key" \
              "&expiration=2099-01-01T00:00:00.000Z" \
              "&signature=weewerwerer23r2"

        # Mock return from create_expiration_date()
        fun.return_value = "2099-01-01T00:00:00.000Z"
        parsed = urlparse(url)
        query_parameters = dict(parse_qsl(parsed.query))
        GATEWAY.verify_query_parameters(query_parameters, 'access_key')
    assert excinfo.type == SignatureInvalidException


@patch.object(Signer, "create_expiration_date")
def test_verify_query_params_no_access_no_secret(fun):
    url = "https://respondent.qa-rex.dynata.com/start" \
          "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
          "&language=es" \
          "&expiration=2099-01-01T00:00:00.000Z" \
          "&signature=e3bc08b50ac9aa261255828" \
          "f741467a956c40c4ce1f0e29bb1bcd23b81f31959"

    # Mock return from create_expiration_date()
    fun.return_value = "2099-01-01T00:00:00.000Z"
    parsed = urlparse(url)
    query_parameters = dict(parse_qsl(parsed.query))

    assert GATEWAY.verify_query_parameters(query_parameters)


@patch.object(Signer, "create_expiration_date")
def test_verify_query_params_secret_and_access_keys(fun):
    url = "https://respondent.qa-rex.dynata.com/start" \
          "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
          "&language=es" \
          "&access_key=access_key" \
          "&secret_key=secret_key" \
          "&expiration=2099-01-01T00:00:00.000Z" \
          "&signature=386b8ad95a284f9e944dd0" \
          "12dfc92c1872790a9bad2e00e19b57c346fb725629"

    # Mock return from create_expiration_date()
    fun.return_value = "2099-01-01T00:00:00.000Z"
    parsed = urlparse(url)
    query_parameters = dict(parse_qsl(parsed.query))

    assert GATEWAY.verify_query_parameters(
        query_parameters,
        'access_key',
        'secret_key'
    )


@patch.object(Signer, "create_expiration_date")
def test_verify_url_success(fun):
    url = "https://respondent.qa-rex.dynata.com/start" \
          "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
          "&language=es" \
          "&access_key=access_key" \
          "&secret_key=secret_key" \
          "&expiration=2099-01-01T00:00:00.000Z" \
          "&signature=386b8ad95a284f9e944dd0" \
          "12dfc92c1872790a9bad2e00e19b57c346fb725629"

    assert GATEWAY.verify_url(url)


@patch.object(Signer, "create_expiration_date")
def test_verify_url_bad_url(fun):
    url = "https://unsigned-url-haha-whattup-hahaha.com/start"

    assert not GATEWAY.verify_url(url)


def test_get_respondent_disposition_complete():
    url = "https://respondent.qa-rex.dynata.com/start" \
          "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
          "&language=es" \
          "&access_key=access_key" \
          "&secret_key=secret_key" \
          "&disposition=1" \
          "&expiration=2099-01-01T00:00:00.000Z" \
          "&signature=386b8ad95a284f9e944dd0" \
          "12dfc92c1872790a9bad2e00e19b57c346fb725629"

    assert GATEWAY.get_respondent_disposition(url) == GatewayDispositionsEnum['COMPLETE']  # noqa


def test_get_respondent_disposition_does_not_exist():
    with raises(ValueError) as excinfo:
        url = "https://respondent.qa-rex.dynata.com/start" \
              "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
              "&language=es" \
              "&access_key=access_key" \
              "&secret_key=secret_key" \
              "&disposition=15" \
              "&expiration=2099-01-01T00:00:00.000Z" \
              "&signature=386b8ad95a284f9e944dd01" \
              "2dfc92c1872790a9bad2e00e19b57c346fb725629"
        GATEWAY.get_respondent_disposition(url)
    assert "15 is not a valid GatewayDispositionsEnum" in str(excinfo.value)


def test_get_respondent_disposition_missing_disposition():
    url = "https://respondent.qa-rex.dynata.com/start" \
          "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
          "&language=es" \
          "&access_key=access_key" \
          "&secret_key=secret_key" \
          "&expiration=2099-01-01T00:00:00.000Z" \
          "&signature=386b8ad95a284f9e944dd01" \
          "2dfc92c1872790a9bad2e00e19b57c346fb725629"
    assert not GATEWAY.get_respondent_disposition(url)


def test_get_respondent_status_missing_disposition():
    url = "https://respondent.qa-rex.dynata.com/start" \
          "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
          "&language=es" \
          "&access_key=access_key" \
          "&secret_key=secret_key" \
          "&expiration=2099-01-01T00:00:00.000Z" \
          "&signature=386b8ad95a284f9e944dd01" \
          "2dfc92c1872790a9bad2e00e19b57c346fb725629"
    assert not GATEWAY.get_respondent_status(url)


def test_get_respondent_status_missing_status():
    url = "https://respondent.qa-rex.dynata.com/start" \
          "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
          "&language=es" \
          "&access_key=access_key" \
          "&secret_key=secret_key" \
          "&disposition=2" \
          "&expiration=2099-01-01T00:00:00.000Z" \
          "&signature=386b8ad95a284f9e944dd01" \
          "2dfc92c1872790a9bad2e00e19b57c346fb725629"
    assert not GATEWAY.get_respondent_status(url)


def test_get_respondent_status():
    url = "https://respondent.qa-rex.dynata.com/start" \
          "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
          "&language=es" \
          "&access_key=access_key" \
          "&secret_key=secret_key" \
          "&disposition=1" \
          "&status=0" \
          "&expiration=2099-01-01T00:00:00.000Z" \
          "&signature=386b8ad95a284f9e944dd01" \
          "2dfc92c1872790a9bad2e00e19b57c346fb725629"

    assert GATEWAY.get_respondent_status(url) == GatewayStatusEnum.COMPLETE_DEFAULT  # noqa


def test_get_respondent_status_does_not_exist():
    with raises(ValueError) as excinfo:
        url = "https://respondent.qa-rex.dynata.com/start" \
              "?ctx=7c26bf58-43db-4370-977d-d14fa4356930" \
              "&language=es" \
              "&access_key=access_key" \
              "&secret_key=secret_key" \
              "&disposition=1" \
              "&status=15" \
              "&expiration=2099-01-01T00:00:00.000Z" \
              "&signature=386b8ad95a284f9e944dd012d" \
              "fc92c1872790a9bad2e00e19b57c346fb725629"
        print(GATEWAY.get_respondent_status(url))
    assert "is not a valid GatewayStatusEnum" in str(excinfo.value)


@patch.object(RexRequest, "post")
def test_create_context(fun):
    expected = 12345

    # Mock return from create_expiration_date()
    fun.return_value = {'id': 12345}

    context = GATEWAY.create_context(
        "12345",
        {
            "ctx": "a987dsglh34t435jkhsdg98u",
            "gender": "male",
            "postal_code": "60081",
            "birth_date": "1959-10-05",
            "country": "US"
        }
    )
    assert expected == context


@patch.object(requests.Session, "post")
def test_expire_context(fun):
    fun.return_value = ResponseMock._response_mock(
        200,
        content_type="text/plain"
    )

    assert GATEWAY.expire_context("12345") is None


@patch.object(requests.Session, "post")
def test_get_context(fun):
    expected = {
        "id": "string",
        "expiration": "2019-08-24T14:15:22Z",
        "items": {
            "property1": "string",
            "property2": "string"
        }
    }
    fun.return_value = ResponseMock._response_mock(
        200,
        content=json.dumps({
            "id": "string",
            "expiration": "2019-08-24T14:15:22Z",
            "items": {
                "property1": "string",
                "property2": "string"
            }
        }),
        content_type="text/plain"
    )

    context = GATEWAY.get_context(12345)
    assert context == expected


@patch.object(requests.Session, "post")
def test_list_attributes(fun):
    expected = {
        "data": [
            {
                "active": "bool",
                "parameter_id": "int"
            }
        ]
    }
    fun.return_value = ResponseMock._response_mock(
        200,
        content=json.dumps({
            "data": [
                {
                    "active": "bool",
                    "parameter_id": "int"
                }
            ]
        }),
        content_type="text/plain"
    )
    context = GATEWAY.list_attributes("US", 0, 100)
    assert context == expected


@patch.object(requests.Session, "post")
def test_get_attribute_info(fun):
    expected = {
        "id": "int",
        "name": "string",
        "description": "string",
        "display_mode": "string",
        "parent_dependencies": [
            {
                "answer_ids": [
                    {
                        "id": "int"
                    }
                ],
                "parameter_id": "int"
            }
        ],
        "expiration_duration": "int",
        "is_active": "bool",
        "countries": [
            {
                "code": "string"
            }
        ],
        "question": {
            "text": "string",
            "translations": [
                {
                    "locale": "string",
                    "text": "string"
                }
            ]
        },
        "answers": [
            {
                "id": "int",
                "text": "string",
                "countries": [
                    {
                        "code": "string"
                    }
                ],
                "translations": [
                    {
                        "locale": "string",
                        "text": "string"
                    }
                ]
            }
        ]
    }
    fun.return_value = ResponseMock._response_mock(
        200,
        content=json.dumps({
            "id": "int",
            "name": "string",
            "description": "string",
            "display_mode": "string",
            "parent_dependencies": [
                {
                    "answer_ids": [
                        {
                            "id": "int"
                        }
                    ],
                    "parameter_id": "int"
                }
            ],
            "expiration_duration": "int",
            "is_active": "bool",
            "countries": [
                {
                    "code": "string"
                }
            ],
            "question": {
                "text": "string",
                "translations": [
                    {
                        "locale": "string",
                        "text": "string"
                    }
                ]
            },
            "answers": [
                {
                    "id": "int",
                    "text": "string",
                    "countries": [
                        {
                            "code": "string"
                        }
                    ],
                    "translations": [
                        {
                            "locale": "string",
                            "text": "string"
                        }
                    ]
                }
            ]
        }),
        content_type="text/plain"
    )
    context = GATEWAY.get_attribute_info(402)
    assert context == expected


@patch.object(requests.Session, 'post')
def test_put_respondent_answers(fun):
    expected = TEST_DATA['test_put_respondent_answers']

    fun.return_value = ResponseMock._response_mock(
        200,
        content=json.dumps(expected),
        content_type="text/plain"
    )

    context = GATEWAY.put_respondent_answers(
        PutRespondentAnswersRequest(
            "123456789",
            [Attribute(12, [4, 5, 3, 2])]
        )
    )
    assert context == expected



from dynata_rex import OpportunityRegistry
registry = OpportunityRegistry('rex_access_key', 'rex_secret_key')

# List opportunities from the registry
opportunities = registry.list_opportunities()

opportunity = opportunities[0]

opportunity_json = opportunity.json()



from dynata_rex import RespondentGateway
gateway = RespondentGateway('rex_access_key', 'rex_secret_key')


############################################
# Get Disposition of a Survey from Endlink #
############################################

termination = "https://respondent.fake.rex.dynata.com/end?ctx=XXXX&transaction_id=123456&disposition=2&status=1&access_key=rex_access_key&expiration=2021-11-24T19:23:23.439Z&signature=d351ff102b3ae6252d47fd54b859ecaf38c2701f214c233848bbdf64c0bc7fe1"  # noqa

disposition = gateway.get_respondent_disposition(termination)

# >>> <GatewayDispositionsEnum.TERMINATION: 2>

disposition.name

# >>> 'TERMINATION'

disposition.value

# >>> 2


#####################################################
# Get Disposition + Status of a Survey from Endlink #
#####################################################

termination = "https://respondent.fake.rex.dynata.com/end?ctx=XXXX&transaction_id=123456&disposition=2&status=1&access_key=rex_access_key&expiration=2021-11-24T19:23:23.439Z&signature=d351ff102b3ae6252d47fd54b859ecaf38c2701f214c233848bbdf64c0bc7fe1"  # noqa

status = gateway.get_respondent_status(termination)

# <GatewayStatusEnum.TERMINATION_DYNATA: (<GatewayDispositionsEnum.TERMINATION: 2>, 1)>  # noqa

status.name

# >>> 'TERMINATION_DYNATA'

status.value

# >>> (<GatewayDispositionsEnum.TERMINATION: 2>, 1)



from dynata_rex import RespondentGateway
gateway = RespondentGateway('rex_access_key', 'rex_secret_key')

url = 'https://respondent.fake.rex.dynata.com/start?ctx=XXXX&language=en'
date_of_birth = '1990-01-01'
gender = 'male'
postal_code = '90210'
respondent_id = 'very-unique-respondent-id'
ttl = 10  # in seconds

custom_params = {
    'custom_parameter': 'custom_value',
    'another_custom_parameter': 'another_custom_value'
}

signed_link = gateway.create_respondent_url(url,
                                            date_of_birth,
                                            gender,
                                            postal_code,
                                            respondent_id,
                                            additional_params=custom_params,
                                            ttl=ttl)

# https://respondent.fake.rex.dynata.com/start?ctx=XXXX&language=en&custom_parameter=custom_value&another_custom_parameter=another_custom_value&birth_date=1990-01-01&gender=male&postal_code=90210&respondent_id=very-unique-respondent-id&access_key=rex_access_key&expiration=2021-12-02T13:48:55.759Z&signature=cf443326b73fb8af14c590e18d79a970fc3f73327c2d140c324ee1ce3020d064  # noqa



"""
Package: src.tests
Filename: test_requester.py
Author(s): Grant W

Description: Test our requester

NOTE: Requester mainly uses a TTL (seconds from current time), so we have to
mock the expiration date generated quite a bit in here
"""
# Python Imports
from unittest.mock import patch
import pytest
import json

# Third Party Imports
import requests

# Dynata Imports
import dynata_rex
from dynata_rex.signer import RexRequest, Signer

# Local Imports
from .shared import (ACCESS_KEY,
                     SECRET_KEY,
                     DEFAULT_PARAMETERS,
                     TEST_DATE_STR,
                     ResponseMock)

REQUESTER = RexRequest(ACCESS_KEY, SECRET_KEY)


@patch.object(Signer, "create_expiration_date")
def test_create_auth_headers(fun):
    # Mock return from create_expiration_date()
    fun.return_value = TEST_DATE_STR
    signature = \
        '6aa2d187875966c2c75fd2cd4db5bbb2de4bafab2494ced8154e3c3b0a906006'
    signing_string = \
        'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    expect = {
        'dynata-expiration': TEST_DATE_STR,
        'dynata-access-key': ACCESS_KEY,
        'dynata-signature': signature,
        'dynata-signing-string': signing_string
    }
    headers = REQUESTER._create_auth_headers()

    for key, val in headers.items():
        assert key in expect.keys()
        assert expect[key] == val


@patch.object(Signer, "create_expiration_date")
def test_create_auth_headers_additional_params(fun):
    # Mock return from create_expiration_date()
    fun.return_value = TEST_DATE_STR
    signature = \
        '6aa2d187875966c2c75fd2cd4db5bbb2de4bafab2494ced8154e3c3b0a906006'
    signing_string = \
        'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    base = {
        'dynata-expiration': TEST_DATE_STR,
        'dynata-access-key': ACCESS_KEY,
        'dynata-signature': signature,
        'dynata-signing-string': signing_string
    }
    expect = dict(base, **DEFAULT_PARAMETERS)
    headers = REQUESTER._create_auth_headers(
        additional_headers=DEFAULT_PARAMETERS)

    for key, val in headers.items():
        assert key in expect.keys()
        assert expect[key] == val


@patch.object(requests.Session, "get")
def test_get_504_raises_timeout_exception(fun):

    fun.return_value = ResponseMock._504()

    with pytest.raises(dynata_rex.exceptions.HttpTimeoutException):
        REQUESTER.get('https://not-a-real-url-abcdefg.com')


@patch.object(requests.Session, "post")
def test_post_504_raises_timeout_exception(fun):

    fun.return_value = ResponseMock._504()

    with pytest.raises(dynata_rex.exceptions.HttpTimeoutException):
        REQUESTER.post('https://not-a-real-url-abcdefg.com', data='{}')


@patch.object(requests.Session, "get")
def test_get_5XX_raises_service_exception(fun):

    status_codes = list(range(400, 600))
    status_codes.remove(504)  # 504 is handled differently

    for status_code in status_codes:
        fun.return_value = ResponseMock._response_mock(status_code)
        with pytest.raises(dynata_rex.exceptions.RexServiceException):
            REQUESTER.get('https://not-a-real-url-abcdefg.com')


@patch.object(requests.Session, "post")
def test_post_5XX_raises_service_exception(fun):

    status_codes = list(range(400, 600))
    status_codes.remove(504)  # 504 is handled differently

    for status_code in status_codes:
        fun.return_value = ResponseMock._response_mock(status_code)
        with pytest.raises(dynata_rex.exceptions.RexServiceException):
            REQUESTER.post('https://not-a-real-url-abcdefg.com', data='{}')


@patch.object(requests.Session, "get")
def test_get_returns_dict_from_json(fun):

    fun.return_value = ResponseMock._response_mock(
        200,
        content=json.dumps({"whoo": "hoo"}),
        content_type="application/json"
    )

    r = REQUESTER.get('https://not-a-real-url-abcdefg.com')
    assert isinstance(r, dict)


@patch.object(requests.Session, "post")
def test_post_returns_dict_from_json(fun):

    fun.return_value = ResponseMock._response_mock(
        200,
        content=json.dumps({"whoo": "hoo"}),
        content_type="application/json"
    )

    r = REQUESTER.post('https://not-a-real-url-abcdefg.com', data='{}')
    assert isinstance(r, dict)


@patch.object(requests.Session, "get")
def test_get_returns_content_from_non_json(fun):

    fun.return_value = ResponseMock._response_mock(
        200,
        content="hello",
        content_type="text/plain"
    )

    r = REQUESTER.get('https://not-a-real-url-abcdefg.com')
    assert isinstance(r, str)


@patch.object(requests.Session, "post")
def test_post_returns_content_from_non_json(fun):

    fun.return_value = ResponseMock._response_mock(
        200,
        content="hello",
        content_type="text/plain"
    )

    r = REQUESTER.post('https://not-a-real-url-abcdefg.com', data='{}')
    assert isinstance(r, str)



"""
Package: dynata-rex.models
Filename: opportunity_registry.py
Author(s): Grant W

Description: Implementation of dataclasses for Opportunity Registry opps
"""
# Python Imports
from enum import Enum
from typing import List, Optional, Union
from typing_extensions import Literal

# Third Party Imports
from pydantic import Field, HttpUrl

# Local Imports
from .base import HashableModel, FallbackEnum


class CategoryEnum(FallbackEnum):
    DEFAULT = 0
    AUTOMOTIVE = 550
    BEVERAGES_ALCOHOLIC = 551
    BEVERAGES_NON_ALCOHOLIC = 552
    BUSINESS = 553
    CHILDREN_PARENTING = 554
    LOYALTY_PROGRAMS = 555
    DESTINATIONS_TOURISM = 556
    EDUCATION = 557
    ELECTRICS_COMPUTERS = 558
    ENTERTAINMENT_LEISURE = 559
    FASHION_CLOTHING = 560
    FINANCE = 561
    FOOD = 562
    GAMBLING_LOTTERY = 563
    GOVERNMENT_POLITICS = 564
    HEALTHCARE = 565
    HOME = 566
    MEDIA_PUBLISHING = 567
    PERSONAL_CARE = 568
    RESTAURANTS = 569
    SENSITIVE_CONTENT = 570
    SMOKING = 571
    SOCIAL_RESEARCH = 572
    SPORTS_FITNESS = 573
    TELECOMMUNICATIONS = 574
    TRANSPORTATION = 575
    TRAVEL_AIRLINES = 576
    TRAVEL_HOTELS = 577
    TRAVEL_SERVICES = 578
    CREDIT_CARDS = 581
    VIDEO_GAMES = 582
    FASHION_CLOTHING_DEPARTMENT_STORE = 583
    ADHOC = 587


class StatusEnum(Enum):
    OPEN = "OPEN"
    PAUSED = "PAUSED"
    CLOSED = "CLOSED"


class EvaluationEnum(Enum):
    STARTS = "STARTS"
    COMPLETES = "COMPLETES"


class DevicesEnum(Enum):
    MOBILE = "mobile"
    DESKTOP = "desktop"
    TABLET = "tablet"


class CellTypeEnum(Enum):
    VALUE = "VALUE"
    RANGE = "RANGE"
    LIST = "LIST"
    COLLECTION = "COLLECTION"


class Locale(HashableModel):
    language: str
    country: str


class Range(HashableModel):
    from_: Union[int, None] = Field(alias='from')
    to: Union[int, None] = None


class Cell(HashableModel):
    tag: str
    attribute_id: int
    negate: bool
    kind: CellTypeEnum


class RangeCell(Cell):
    kind: Literal[CellTypeEnum.RANGE.value]
    range_: Range = Field(alias='range')


class ValueCell(Cell):
    kind: Literal[CellTypeEnum.VALUE.value]
    value: str = None


class ListCell(Cell):
    kind: Literal[CellTypeEnum.LIST.value]
    list_: List[str] = Field(alias='list')


class CollectionCell(Cell):
    kind: Literal[CellTypeEnum.COLLECTION.value]
    collection_: str = Field(alias='collection')


class Links(HashableModel):
    live: HttpUrl
    # test: HttpUrl


class Quota(HashableModel):
    id: str
    cells: List[str]
    count: int
    status: StatusEnum


class Filter(HashableModel):
    id: str
    cells: List[str]


class Opportunity(HashableModel):
    class Config:
        allow_population_by_field_name = True
        validate_assignment = True
        allow_mutation = False

    id: int
    status: StatusEnum
    length_of_interview: int
    incidence_rate: int
    cost_per_interview: float
    completes: int
    project_id: int
    group_id: int
    evaluation: EvaluationEnum
    days_in_field: int
    locale: Locale
    links: Links
    client_id: Optional[Union[None, int]]

    project_exclusions: List[int]
    category_exclusions: List[CategoryEnum]
    category_ids: List[CategoryEnum]
    devices: List[DevicesEnum]
    filters: List[List[Filter]]
    cells: List[Union[RangeCell, ListCell, ValueCell, CollectionCell]]
    quotas: List[List[Quota]]


class Invite(HashableModel):
    id: int
    collection_id: str
    respondent_id: str
    expiration: Optional[int]
    modified: str
    created: str



"""
Package: src.smor
Filename: api.py
Author(s): Grant W

Description: API interaction for smor
"""
# Python Imports
import json
from typing import List

# Third Party Imports
import pydantic

# Local Imports
import dynata_rex.models as models
from .signer import RexRequest
from .logs import logger
from .exceptions import InvalidShardException


class OpportunityRegistry:
    _BASE_URL = 'https://registry.rex.dynata.com'

    def __init__(self,
                 access_key: str,
                 secret_key: str,
                 base_url: str = _BASE_URL,
                 default_ttl: int = 10,
                 shard_count: int = 1,
                 current_shard: int = 1):
        """
        @access_key: liam access key for REX
        @secret_key: liam secret key for REX

        # Optional
        @base_url  : url of Opportunity Registry
        @shard_count  : number of total shards consuming Opportunity Registry
        @current_shard: curent shard
        @default_ttl: time to live for signature in seconds
        """
        self.default_ttl = default_ttl
        self.make_request = RexRequest(access_key,
                                       secret_key,
                                       default_ttl=default_ttl)
        self.base_url = self._format_base_url(base_url)

        if current_shard > shard_count:
            raise InvalidShardException
        self.shard_count = shard_count
        self.current_shard = current_shard

    def _format_base_url(self, url):
        """Make sure our URL starts with http, and if not add https://
        prefix"""
        if url.startswith('http'):
            return url
        return f"https://{url}"

    def _get_opportunity(self, opportunity_id: int) -> dict:
        """Raw get opportunity"""
        endpoint = f"{self.base_url}/get-opportunity"
        data = {"id": opportunity_id}
        return self.make_request.post(endpoint, data)

    def _list_opportunities(self, limit: int = 10) -> List[dict]:
        """
        [Deprecated - please use _receive_notifications()]
        Raw get opportunities"""
        endpoint = f"{self.base_url}/list-opportunities"
        data = {
            "limit": limit,
            "shards": {
                "count": self.shard_count,
                "current": self.current_shard
            }
        }
        return self.make_request.post(endpoint, data)

    def _receive_notifications(self, limit: int = 10) -> List[dict]:
        """Raw receive notifications"""
        endpoint = f"{self.base_url}/receive-notifications"
        data = {
            "limit": limit,
            "shards": {
                "count": self.shard_count,
                "current": self.current_shard
            }
        }
        return self.make_request.post(endpoint, data)

    def _receive_invites(self, limit: int = 10) -> list[dict]:
        """Raw receive invites"""
        endpoint = f"{self.base_url}/receive_invites"
        data = {
            "limit": limit
        }
        return self.make_request.post(endpoint, data)

    def list_opportunities(self, limit: int = 10) -> List[models.Opportunity]:
        """
        [Deprecated - please use receive_notifications()]
        Get opportunities from Opportunity Registry"""
        opportunities = self._list_opportunities(limit=limit)
        out = []
        for opp in opportunities:
            try:
                out.append(models.Opportunity(**opp))
            except pydantic.error_wrappers.ValidationError:
                opportunity_id = opp['id']
                logger.warning(
                    f"Unable to parse {opportunity_id}, excluding...")
                logger.warning(json.dumps(opp, indent=4))
                # Ack opportunity so we don't see it again
                self.ack_opportunity(opportunity_id)

        return out

    def receive_notifications(self,
                              limit: int = 10) -> List[models.Opportunity]:
        """Get opportunity notifications from Opportunity Registry"""
        opportunities = self._receive_notifications(limit=limit)
        out = []
        for opp in opportunities:
            try:
                out.append(models.Opportunity(**opp))
            except pydantic.error_wrappers.ValidationError:
                opportunity_id = opp['id']
                logger.warning(
                    f"Unable to parse {opportunity_id}, excluding...")
                logger.warning(json.dumps(opp, indent=4))
                # Ack notification so we don't see it again
                self.ack_notification(opportunity_id)

        return out

    def get_opportunity(self, opportunity_id: int) -> models.Opportunity:
        """Get specific opportunity from SMOR
        """
        opportunity = self._get_opportunity(opportunity_id)
        return models.Opportunity(**opportunity)

    def list_project_opportunities(self, project_id: int) -> List[int]:
        """List related opportunities from a project id"""
        endpoint = f"{self.base_url}/list-project-opportunities"
        data = {"project_id": project_id}
        return self.make_request.post(endpoint, data)

    def ack_opportunity(self, opportunity_id: int) -> None:
        """
        [Deprecated - please use ack_notification()]
        Acknowledge a single opportunity"""
        data = [opportunity_id]
        return self.ack_opportunities(data)

    def ack_notification(self, opportunity_id: int) -> None:
        """Acknowledge a single notification"""
        data = [opportunity_id]
        return self.ack_notifications(data)

    def ack_opportunities(self, opportunities: List[int]) -> None:
        """
        [Deprecated - please use ack_notifications()]
        Acknowledge a list of opportunities"""
        endpoint = f"{self.base_url}/ack-opportunities"
        res = self.make_request.post(endpoint, opportunities)
        return res

    def ack_notifications(self, opportunities: List[int]) -> None:
        """Acknowledge a list of notifications"""
        endpoint = f"{self.base_url}/ack-notifications"
        res = self.make_request.post(endpoint, opportunities)
        return res

    def download_collection(self, collection_id: str) -> list:
        """Download targeting from a collection cell"""
        endpoint = f"{self.base_url}/download-collection"
        data = {"id": str(collection_id)}
        res = self.make_request.post(endpoint, data)
        return str(res).split('\n')

    def ack_invites(self, invites: list[int]) -> None:
        """Acknowledge a list of invites"""
        endpoint = f"{self.base_url}/ack-invites"
        res = self.make_request.post(endpoint, invites)
        return res

    def receive_invites(self, limit: int = 10) -> List[models.Invite]:
        """Receive invites from opportunity registry"""
        invites = self._receive_invites(limit=limit)
        out = []
        for inv in invites:
            try:
                out.append(models.Invite(**inv))
            except pydantic.error_wrappers.ValidationError:
                invite_id = inv['id']
                logger.warning(
                    f"Unable to parse {invite_id}, excluding...")
                logger.warning(json.dumps(inv, indent=4))
                # Ack invites so we don't see it again
                self.ack_invites(invite_id)
        return out

    def download_invite_collection(self, invite_id: str) -> list:
        """Download invite collection from opportunity registry"""
        endpoints = f"{self.base_url}/download-invite-collection"
        data = {"id": invite_id}
        res = self.make_request.post(endpoints, data)
        return str(res).split('\n')



from dynata_rex import RespondentGateway
gateway = RespondentGateway('rex_access_key', 'rex_secret_key')

attribute_id = 123  # Some-Number

gateway.get_attribute_info(attribute_id)

# {
#     "id": "unique-parameter-id",
#     "name": "Example Question Name",
#     "description": "Example Question Description",
#     "display_mode": "Single-Punch/Multi-Punch",
#     "parent_dependencies": [
#         {
#             "answer_ids": [
#                 {
#                     "id": "unique-answer-id"
#                 }
#             ],
#             "parameter_id": "unique-parameter-id"
#         }
#     ],
#     "expiration_duration": "36000",
#     "is_active": "true/false",
#     "countries": [
#         {
#             "code": "country-code"
#         }
#     ],
#     "question": {
#         "text": "question-text",
#         "translations": [
#             {
#                 "locale": "locale",
#                 "text": "translation-text"
#             }
#         ]
#     },
#     "answers": [
#         {
#             "id": "unique-answer-id",
#             "text": "answer-text",
#             "countries": [
#                 {
#                     "code": "country-code"
#                 }
#             ],
#             "translations": [
#                 {
#                     "locale": "locale",
#                     "text": "translation-text"
#                 }
#             ]
#         }
#     ]
# }



"""
Package: src.tests
Filename: test_opportunity_registry.py
Author(s): Grant W

Description: Tests for the opportunity registry
"""
# Python Imports
from unittest.mock import patch
import json

# Third Party Imports
import requests
import pytest

# Dynata Imports
import dynata_rex

import dynata_rex.models

# Local Imports
from .shared import (ACCESS_KEY,
                     SECRET_KEY,
                     BASE_URL,
                     ResponseMock,
                     TEST_DATA)

REGISTRY = dynata_rex.OpportunityRegistry(ACCESS_KEY, SECRET_KEY, BASE_URL)


def test_format_base_url_adds_https_scheme():
    fake_url = 'missing_fake_base_url.com'
    url = REGISTRY._format_base_url(fake_url)
    assert url == f'https://{fake_url}'


def test_invalid_shard_current_greater_than_total():
    with pytest.raises(dynata_rex.exceptions.InvalidShardException):
        _ = dynata_rex.OpportunityRegistry(
            ACCESS_KEY,
            SECRET_KEY,
            shard_count=1,
            current_shard=2
        )


@patch.object(requests.Session, "post")
def test__get_opportunity(session_post):
    """_get_opportunity should return the json data from response as a dict"""
    data = TEST_DATA['test_get_opportunity']
    session_post.return_value = ResponseMock._response_mock(
        200, content=json.dumps(data), content_type="application/json"
    )

    r = REGISTRY._get_opportunity('1234567')

    assert isinstance(r, dict)


@patch.object(requests.Session, "post")
def test_get_opportunity(session_post):
    """get_opportunity should return the json data from response loaded
    into an Opportunity object"""
    data = TEST_DATA['test_get_opportunity']
    session_post.return_value = ResponseMock._response_mock(
        200, content=json.dumps(data), content_type="application/json"
    )

    r = REGISTRY.get_opportunity('1234567')

    assert isinstance(r, dynata_rex.models.Opportunity)


@patch.object(requests.Session, "post")
def test__receive_notifications(session_post):
    """_receive_notifications should return the json data from response as a
    list of dicts"""
    data = TEST_DATA['test_receive_notifications']
    session_post.return_value = ResponseMock._response_mock(
        200, content=json.dumps(data), content_type="application/json"
    )

    r = REGISTRY._receive_notifications()

    assert isinstance(r, list)

    for opportunity in r:
        assert isinstance(opportunity, dict)


@patch.object(requests.Session, "post")
def test_receive_notifications(session_post):
    """_receive_notifications should return the json data from response as a
    list of Opportunity objects"""
    data = TEST_DATA['test_receive_notifications']
    session_post.return_value = ResponseMock._response_mock(
        200, content=json.dumps(data), content_type="application/json"
    )
    r = REGISTRY.receive_notifications()

    assert isinstance(r, list)

    for opportunity in r:
        assert isinstance(opportunity, dynata_rex.models.Opportunity)


@patch.object(dynata_rex.OpportunityRegistry, "ack_notification")
@patch.object(requests.Session, "post")
def test_receive_notifications_assert_ack_for_invalid_opportunity(session_post,
                                                                  ack_method):
    """receive notifications should 'ack' a notification returned
    that it cannot convert into an Opportunity object"""
    data = TEST_DATA['test_receive_notifications']

    # Append an invalid Opportunity
    data.append(
        {
            "id": 999999,
            "status": "CLOSED",
            "client_id": None,
            "length_of_interview": 10,
            "invalid_field": "invalid_value"
        }
    )
    session_post.return_value = ResponseMock._response_mock(
        200, content=json.dumps(data), content_type="application/json"
    )
    REGISTRY.receive_notifications()

    # Make sure the ack method was called for the invalid opportunity
    assert ack_method.call_count == 1


@patch.object(dynata_rex.OpportunityRegistry, "ack_notifications")
@patch.object(requests.Session, "post")
def test_ack_notification(session_post, ack_method):
    data = TEST_DATA['test_receive_notifications'][0]

    session_post.return_value = ResponseMock._response_mock(
        200, content=json.dumps(data), content_type="application/json"
    )
    REGISTRY.ack_notification(data['id'])

    assert ack_method.call_count == 1

    # Make sure we called parent `ack_notifications` method with
    # [ data['id'] ]
    assert ack_method.call_args == (([data['id']],),)


@patch.object(requests.Session, "post")
def test_ack_notifications(session_post):
    """receive notifications should 'ack' an opportunity returned
       that it cannot convert into an Opportunity object"""

    data = [x['id'] for x in TEST_DATA['test_receive_notifications']]

    session_post.return_value = ResponseMock._response_mock(
        204, content_type="application/json"
    )

    REGISTRY.ack_notifications(data)

    assert session_post.call_count == 1
    assert session_post.call_args[1]['data'] == json.dumps(data)


@patch.object(requests.Session, "post")
def test_list_project_opportunities(session_post):
    data = [17039, 17344, 17038, 17040, 17041, 17042]

    session_post.return_value = ResponseMock._response_mock(
        200, content=json.dumps(data), content_type="application/json"
    )
    res = REGISTRY.list_project_opportunities(99999)

    assert session_post.call_args[1]['data'] == json.dumps({
        'project_id': 99999,
    })
    assert res == data


@patch.object(requests.Session, "post")
def test_download_collection(session_post):
    """list opportunities should 'ack' an opportunity returned
       that it cannot convert into an Opportunity object"""

    data = TEST_DATA['test_download_collection']

    session_post.return_value = ResponseMock._response_mock(
        200, content=data, content_type="text/csv"
    )

    r = REGISTRY.download_collection('1234567')

    assert isinstance(r, list)


@patch.object(requests.Session, "post")
def test_download_invite_collection(session_post):
    data = TEST_DATA['test_download_invite_collection']

    session_post.return_value = ResponseMock._response_mock(
        200, content=data, content_type="text/csv"
    )

    r = REGISTRY.download_invite_collection("12345567")

    assert isinstance(r, list)


@patch.object(requests.Session, "post")
def test__receive_invites(session_post):
    data = TEST_DATA['test_receive_invites']

    session_post.return_value = ResponseMock._response_mock(
        200, content=json.dumps(data), content_type="application/json"
    )

    r = REGISTRY._receive_invites()

    assert isinstance(r, list)
    for invite in r:
        assert isinstance(invite, dict)


@patch.object(requests.Session, "post")
def test_receive_invites(session_post):
    data = TEST_DATA['test_receive_invites']

    session_post.return_value = ResponseMock._response_mock(
        200, content=json.dumps(data), content_type="application/json"
    )

    r = REGISTRY.receive_invites()

    assert isinstance(r, list)

    for invite in r:
        assert isinstance(invite, dynata_rex.models.Invite)



"""
Package: dynata-rex
Filename: exceptions.py
Author(s): Grant W

Description: Exceptions for the dynata-rex package.
"""
# Python Imports

# Third Party Imports

# Local Imports


class RexClientException(Exception):
    """
    Base Exception for all RexClient exceptions.
    """
    pass


class RexServiceException(RexClientException):
    """
    Base Exception for all RexServer exceptions.
    """
    pass


class InvalidShardException(RexClientException):
    pass


class HttpTimeoutException(RexServiceException):
    pass


class InvalidCredentialsException(RexServiceException):
    pass


class SignatureExpiredException(RexClientException):
    pass


class SignatureInvalidException(RexClientException):
    pass



from .base import (
    HashableModel,
    FallbackEnum
)


from .opportunity_registry import (
    CategoryEnum,
    StatusEnum,
    EvaluationEnum,
    DevicesEnum,
    CellTypeEnum,
    Locale,
    Range,
    Cell,
    RangeCell,
    ValueCell,
    ListCell,
    Links,
    Quota,
    Filter,
    Opportunity,
    Invite
)

from .respondent_gateway import (
    GatewayGenderEnum,
    GatewayDispositionsEnum,
    GatewayStatusEnum,
    PutRespondentRequest,
    PutRespondentAnswersRequest,
    Attribute
)


__all__ = [
    'HashableModel',
    'FallbackEnum',
    'CategoryEnum',
    'StatusEnum',
    'EvaluationEnum',
    'DevicesEnum',
    'CellTypeEnum',
    'Locale',
    'Range',
    'Cell',
    'RangeCell',
    'ValueCell',
    'ListCell',
    'Links',
    'Quota',
    'Filter',
    'Opportunity',
    'GatewayGenderEnum',
    'GatewayDispositionsEnum',
    'GatewayStatusEnum',
    'Invite',
    'PutRespondentRequest',
    "PutRespondentAnswersRequest",
    "Attribute"
]



from dynata_rex import RespondentGateway
gateway = RespondentGateway('rex_access_key', 'rex_secret_key')

country = 'country-code'

page_number = 3  # Page Number Being Requested

page_size = 100  # Number of Items Being Requested

gateway.list_attributes(country, page_number, page_size)

# {
#   "data": [
#       {
#           "active": "true/false",
#           "parameter_id": "unique-parameter-id"
#       }
#   ]
# }



from dynata_rex import RespondentGateway

gateway = RespondentGateway('rex_access_key', 'rex_secret_key')

url = 'https://respondent.fake.rex.dynata.com/start?ctx=XXXX&language=en'

signed_url = gateway.sign_url(url, url_quoting=True)

# 'https://respondent.fake.rex.dynata.com/start?ctx=XXXX&language=en&access_key=rex_access_key&expiration=2021-11-24T16%3A12%3A35.991Z&signature=4219cf63406ae429d94dbe9c33027816c264c1e2bf1edbadd2510eb9bf2351c3'



from dynata_rex import RespondentGateway
gateway = RespondentGateway('rex_access_key', 'rex_secret_key')

# Verify a signed /end link URL with your credentials

#############
# Valid URL #
#############
end_url = "https://respondent.fake.rex.dynata.com/end?ctx=XXXX&transaction_id=123456&disposition=2&status=1&access_key=rex_access_key&expiration=2021-11-24T19:23:23.439Z&signature=d351ff102b3ae6252d47fd54b859ecaf38c2701f214c233848bbdf64c0bc7fe1"  # noqa

gateway.verify_url(end_url)

# >>> True

#####################
# Missing Signature #
#####################

missing_signature = "https://respondent.fake.rex.dynata.com/end?ctx=XXXX&transaction_id=123456&disposition=2&status=1&access_key=rex_access_key&expiration=2021-11-24T19:23:23.439Z"  # noqa

gateway.verify_url(missing_signature)

# >>> False

###############################
# Altered Parameters          #
# (Term --> Complete Attempt) #
###############################
# Disposition changed to 1 (from 2) and status to 0 (from 1)

altered_parameters = "https://respondent.fake.rex.dynata.com/end?ctx=XXXX&transaction_id=123456&disposition=1&status=0&access_key=rex_access_key&expiration=2021-11-24T19:23:23.439Z&signature=d351ff102b3ae6252d47fd54b859ecaf38c2701f214c233848bbdf64c0bc7fe1"  # noqa

gateway.verify_url(altered_parameters)

# >>> False



"""
Package: src.tests
Filename: test_signer.py
Author(s): Grant W

Description: Tests for signature
"""
# Python Imports
from unittest.mock import patch

# Third Party Imports

# Dynata Imports
from dynata_rex.signer import Signer

# Local Imports
from .shared import (ACCESS_KEY,
                     SECRET_KEY,
                     SIGNING_STRING,
                     TEST_DATE_STR,
                     DEFAULT_PARAMETERS)

SIGNER = Signer(ACCESS_KEY, SECRET_KEY)


def test_is_expired_static_date():
    date_str = "1970-01-01T00:00:00.000Z"
    assert SIGNER.is_expired(date_str)


def test_is_expired_generated_date():
    date_str = SIGNER.create_expiration_date(ttl=0)
    assert SIGNER.is_expired(date_str)


def test_is_expired_false_static_date():
    # TODO: If you are reading this in 2099 you should change the date
    date_str = "2099-01-01T00:00:00.000Z"
    assert not SIGNER.is_expired(date_str)


def test_is_expired_false_generated_date():
    date_str = SIGNER.create_expiration_date(ttl=1000)
    assert not SIGNER.is_expired(date_str)


@patch.object(Signer, "create_expiration_date")
def test_generate_signature_from_ttl(fun):
    expect = "f42747bca8a7d0f5ad6adb6eca7c1dd87f7af0b8f9612fba80950fea6e4eff35"

    # Mock return from create_expiration_date()
    fun.return_value = TEST_DATE_STR

    sig, exp = SIGNER.sign_from_ttl(ACCESS_KEY, SECRET_KEY, ttl=0)

    assert sig == expect
    assert exp == TEST_DATE_STR


def test_generate_signature_from_date_str():
    expect = "f42747bca8a7d0f5ad6adb6eca7c1dd87f7af0b8f9612fba80950fea6e4eff35"
    sig, exp = SIGNER.sign_from_expiration_date(ACCESS_KEY,
                                                SECRET_KEY,
                                                TEST_DATE_STR)

    assert sig == expect
    assert exp == TEST_DATE_STR


@patch.object(Signer, "create_expiration_date")
def test_generate_signature_from_ttl_with_signing_string(fun):
    expect = "69e7fcb4a34d5141a368f938f44227a4f264032549ccdc62bd3e652c23e4a8c4"

    # Mock return from create_expiration_date()
    fun.return_value = TEST_DATE_STR

    sig, exp = SIGNER.sign_from_ttl(ACCESS_KEY,
                                    SECRET_KEY,
                                    ttl=0,
                                    signing_string=SIGNING_STRING)

    assert sig == expect
    assert exp == TEST_DATE_STR


def test_generate_signature_from_date_str_with_signing_string():
    expect = "69e7fcb4a34d5141a368f938f44227a4f264032549ccdc62bd3e652c23e4a8c4"
    sig, exp = SIGNER.sign_from_expiration_date(ACCESS_KEY,
                                                SECRET_KEY,
                                                TEST_DATE_STR,
                                                signing_string=SIGNING_STRING)

    assert sig == expect
    assert exp == TEST_DATE_STR


def test_create_query_params_signing_string():
    expect = "13198b48d1ed3ae8d8e2d066fc76cd8b731cea99e3d6f30d9c05ef2e107b4a66"
    sha = SIGNER._create_query_params_signing_string(DEFAULT_PARAMETERS)
    assert sha == expect


def test_sign_query_params_from_expiration_date():
    sig = "035c7ea7c2a7660df6bd8b18a8c6a2b83d869b44c0509e0cb425b460cebbaea4"
    expect = (
        "param_1=value_1"
        "&param_2=value_2"
        "&access_key=access_key"
        "&expiration=1970-01-01T00%3A00%3A00.000Z"
        f"&signature={sig}"
    )
    signed = SIGNER.sign_query_params_from_expiration_date(DEFAULT_PARAMETERS,
                                                           TEST_DATE_STR)
    assert signed == expect


def test_sign_query_params_from_expiration_date_as_dict():
    sig = "035c7ea7c2a7660df6bd8b18a8c6a2b83d869b44c0509e0cb425b460cebbaea4"
    expect = {
        "param_1": "value_1",
        "param_2": "value_2",
        "access_key": ACCESS_KEY,
        "expiration": TEST_DATE_STR,
        "signature": sig
    }
    signed = SIGNER.sign_query_params_from_expiration_date(DEFAULT_PARAMETERS,
                                                           TEST_DATE_STR,
                                                           as_dict=True)
    assert isinstance(signed, dict)
    assert signed == expect


@patch.object(Signer, "create_expiration_date")
def test_sign_query_params_from_ttl(fun):
    # Mock return from create_expiration_date()
    fun.return_value = TEST_DATE_STR

    sig = "035c7ea7c2a7660df6bd8b18a8c6a2b83d869b44c0509e0cb425b460cebbaea4"
    expect = (
        "param_1=value_1"
        "&param_2=value_2"
        "&access_key=access_key"
        "&expiration=1970-01-01T00%3A00%3A00.000Z"
        f"&signature={sig}"
    )
    signed = SIGNER.sign_query_params_from_ttl(DEFAULT_PARAMETERS,
                                               TEST_DATE_STR)
    assert signed == expect


@patch.object(Signer, "create_expiration_date")
def test_sign_query_params_from_ttl_as_dict(fun):
    # Mock return from create_expiration_date()
    fun.return_value = TEST_DATE_STR

    sig = "035c7ea7c2a7660df6bd8b18a8c6a2b83d869b44c0509e0cb425b460cebbaea4"
    expect = {
        "param_1": "value_1",
        "param_2": "value_2",
        "access_key": ACCESS_KEY,
        "expiration": TEST_DATE_STR,
        "signature": sig
    }
    signed = SIGNER.sign_query_parameters_from_ttl(DEFAULT_PARAMETERS,
                                                   TEST_DATE_STR,
                                                   as_dict=True)
    assert isinstance(signed, dict)
    assert signed == expect



from dynata_rex import OpportunityRegistry

registry = OpportunityRegistry('rex_access_key', 'rex_secret_key')

opportunities = registry.receive_notifications()

opportunity = opportunities[0]

registry.ack_notification(opportunity.id)



from dynata_rex import OpportunityRegistry

registry = OpportunityRegistry('rex_access_key', 'rex_secret_key')

opportunities = registry.receive_notifications()

# Returns a list of Opportunity notifications
# [Opportunity(id=1,...), Opportunity(id=2,...), Opportunity(id=1,...)]



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()


setuptools.setup(
    name="dynata_rex",
    version="1.3.1",
    author="REX Maintainers",
    author_email="tech.supply@Dynata.com",
    description=("Package for building and interacting with the "
                 "Dynata Respondent Exchange (REX)"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dynata/rex-sdk-python",
    packages=setuptools.find_packages(exclude=('tests', )),
    platforms=['Any'],
    install_requires=requirements,
    setup_requires=['pytest-runner'],
    extras_require={
        # pip install -e ".[testing]"
        "testing": ['pytest'],
        ':python_version == "3.6"': [
            "typing-extensions==4.7.1",
            'dataclasses==0.8'
        ],
        ':python_version == "3.7"': [
            "typing-extensions==4.7.1"
        ]
    },
    tests_require=['pytest'],
    keywords='respondent exchange rex smor dynata python',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



"""
Package: src.dynata_rex
Filename: __init__.py
Author(s): Grant W

Description: Expose the main classes of the package.
"""
# Python Imports

# Third Party Imports

# Local Imports
from .opportunity_registry import OpportunityRegistry
from .respondent_gateway import RespondentGateway
from .exceptions import (
    RexClientException,
    RexServiceException,
    InvalidShardException,
    HttpTimeoutException,
    InvalidCredentialsException
)

__all__ = [
    'RespondentGateway',
    'OpportunityRegistry',
    'RexClientException',
    'RexServiceException',
    'InvalidShardException',
    'HttpTimeoutException',
    'InvalidCredentialsException'
]



from dynata_rex import OpportunityRegistry

registry = OpportunityRegistry('rex_access_key', 'rex_secret_key')

opportunities = registry.receive_notifications()

opportunity_ids = [opportunity.id for opportunity in opportunities]

registry.ack_notifications(opportunity_ids)



"""
Package: src.tests
Filename: test_models_base.py
Author(s): Grant W

Description: Test for the base models we implement
"""
# Python Imports
from typing import List

# Third Party Imports
import pytest

# Dynata Imports
import dynata_rex

# Local Imports


def test_base_object():
    """Test the base object"""

    class TestObject(dynata_rex.models.base.BaseObject):
        pass

    test_object = TestObject()

    assert str(test_object) == "<dynata_rex.TestObject>"


def test_hashable_model():

    class TestModel(dynata_rex.models.base.HashableModel):
        attribute_one: str
        attribute_two: str

    test_model_one = TestModel(attribute_one="one", attribute_two="one")
    test_model_two = TestModel(attribute_one="two", attribute_two="two")

    # Set uses __hash__ to compare objects
    models = set([test_model_one, test_model_two])

    assert len(models) == 2


def test_hashable_model_with_submodel():

    class TestSubModel(dynata_rex.models.base.HashableModel):
        attribute_three: str
        attribute_four: str

    class TestModel(dynata_rex.models.base.HashableModel):
        attribute_one: str
        attribute_two: str
        attribute_three: List[TestSubModel]

    sub_model_one = TestSubModel(attribute_three="one", attribute_four="one")
    sub_model_two = TestSubModel(attribute_three="two", attribute_four="two")

    test_model_one = TestModel(attribute_one="one",
                               attribute_two="one",
                               attribute_three=[sub_model_one])

    test_model_two = TestModel(attribute_one="two",
                               attribute_two="two",
                               attribute_three=[sub_model_two])

    # Set uses __hash__ to compare objects
    models = set([test_model_one, test_model_two])

    assert len(models) == 2


def test_fallback_enum():

    class TestEnum(dynata_rex.models.base.FallbackEnum):
        DEFAULT = "default"

    test_enum = TestEnum('missing_value')
    assert test_enum.value == "default"


def test_fallback_enum_raises_value_error_when_undefined():

    class TestEnum(dynata_rex.models.base.FallbackEnum):
        pass

    with pytest.raises(ValueError):
        _ = TestEnum('missing')
